/*******************************************************************************

  Intel Data Center Bridging (DCB) Software
  Copyright(c) 2007-2009 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

  Contact Information:
  e1000-eedc Mailing List <e1000-eedc@lists.sourceforge.net>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/

#include "includes.h"
#include "common.h"
#include <stdio.h>
#include <syslog.h>
#include <sys/un.h>
#include <sys/stat.h>
#include "dcbd.h"
#include "dcb_types.h"
#include "dcb_events.h"
#include "dcb_protocol.h"
#include "clif_cmds.h"
#include "ctrl_iface.h"
#include "drv_cfg.h"
#include "config.h"

extern u8 gdcbx_subtype;

static char *hexlist = "0123456789abcdef";
static char *hexlistcaps = "0123456789ABCDEF";

static int get_dcb_state(char *port_id, char *rbuf);
static dcb_result set_dcb_state(char *port_id, char *ibuf, int ilen);
static int get_dcbx_config(u8 cmd, char *rbuf);
static dcb_result set_dcbx_config(char *ibuf, int ilen);
static dcb_result get_pg_data(pg_attribs *pg_data, int cmd, char *port_id,
		int dcbx_st, char *rbuf);
static int set_pg_config(pg_attribs *pg_data, char *port_id, char *ibuf,
		int ilen);
static int set_pfc_config(pfc_attribs *pfc_data, char *port_id, char *ibuf,
		int ilen);
static int get_pfc_data(pfc_attribs *pfc_data, int cmd, char *port_id,
		int dcbx_st, char *rbuf);
static int set_app_config(app_attribs *app_data, char *port_id, u32 subtype,
		char *ibuf, int ilen);
static int get_app_data(app_attribs *app_data, int cmd, char *port_id,
		u32 subtype, int dcbx_st, char *rbuf);
static int set_llink_config(llink_attribs *app_data, char *port_id, u32 subtype,
		char *ibuf, int ilen);
static int get_llink_data(llink_attribs *llink_data, int cmd, char *port_id,
		u32 subtype, int dcbx_st, char *rbuf);
static dcb_result get_bwg_desc(char *port_id, char *ibuf, int ilen, char *rbuf);
static dcb_result set_bwg_desc(char *port_id, char *ibuf, int ilen);
static void set_protocol_data(feature_protocol_attribs *protocol, char *ibuf,
	int plen);
static dcb_result get_cmd_protocol_data(feature_protocol_attribs *protocol,
	u8 cmd, char *rbuf);
static int get_bcn_data(bcn_attribs *bcn_data, int cmd, char *port_id, 
						int dcbx_st, char *rbuf);
static int set_bcn_config(bcn_attribs *bcn_data, char *port_id, char *ibuf, 
						  int ilen);

struct clif_cmds {
	int cmd;
	int (*cmd_handler)(struct dcbd_data *dd,
			     struct sockaddr_un *from,
			     socklen_t fromlen,
			     char *ibuf, int ilen,
			     char *rbuf);
};	

const struct clif_cmds cmd_tbl[] = {
	{ DCB_CMD,     clif_iface_dcb_cmd },
	{ ATTACH_CMD,  clif_iface_attach },
	{ DETACH_CMD,  clif_iface_detach },
	{ LEVEL_CMD,   clif_iface_level },
	{ PING_CMD,    clif_iface_ping },
	{ UNKNOWN_CMD, clif_iface_cmd_unknown }
};

static int get_dcb_state(char *port_id, char *rbuf)
{
	int state = 0;
	dcb_result rval = dcb_success;
	
	if (get_hw_state(port_id, &state))
		rval = dcb_failed;
	else
		sprintf(rbuf, "%01x", state);

	return (rval);
}

static dcb_result set_dcb_state(char *port_id, char *ibuf, int ilen)
{
	int state;
	int tmp;
	int off;
	int plen;
	dcb_result rval = dcb_success;

	plen=strlen(port_id);
	off = DCB_PORT_OFF + plen;

	if (ilen == (off + CFG_DCB_DLEN)) {
		state = (*(ibuf+off+DCB_STATE)) ^ '0';
		if (state == FALSE || state == TRUE) {
			/* if get_hw_state fails, then don't bother
			 * trying to set the state.
			*/
			if (get_hw_state(port_id, &tmp) ||
				set_hw_state(port_id, state))
				rval = dcb_failed;
			else
				rval = save_dcb_enable_state(port_id, state);
		} else {
			rval = dcb_bad_params;
		}
	} else {
		printf("error - setcommand has invalid argument length\n");
		rval = dcb_bad_params;
	}

	return rval;
}

static int get_dcbx_config(u8 cmd, char *rbuf)
{
	dcb_result rval = dcb_success;
	int dcbx_version;
	
	if (cmd == CMD_GET_CONFIG) {
		if (!get_dcbx_version(&dcbx_version))
			rval = dcb_failed;
	} else {
		dcbx_version = gdcbx_subtype;
	}

	if (rval == dcb_success)
		sprintf(rbuf, "%01x", dcbx_version);

	return (rval);
}

static dcb_result set_dcbx_config(char *ibuf, int ilen)
{
	int version;
	int off;
	dcb_result rval = dcb_success;

	off = DCBX_CFG_OFF;

	if (ilen == DCBX_CFG_OFF + CFG_DCBX_DLEN) {
		version = (*(ibuf+off+DCBX_VERSION)) ^ '0';
		switch (version) {
		case dcbx_subtype1:
		case dcbx_subtype2:
			rval = save_dcbx_version(version);
			break;
		default:
			rval = dcb_bad_params;
			break;
		}
	} else {
		printf("error - setcommand has invalid argument length\n");
		rval = dcb_bad_params;
	}

	return rval;
}

static dcb_result get_bwg_desc(char *port_id, char *ibuf, int ilen, char *rbuf)
{
	u8 pgid;
	int plen;
	char *desc_str = NULL;
	dcb_result rval = dcb_success;

	plen = strlen(port_id);

	if (ilen !=  DCB_PORT_OFF + plen + PG_DESC_GET_DLEN) {
		rval = dcb_bad_params;
	} else {
		pgid = *(ibuf + DCB_PORT_OFF + plen + PG_DESC_PGID) & 0x0f;

		if (pgid < 1 || pgid > 8) {
			rval = dcb_bad_params;
		} else {
			rval = get_bwg_descrpt(port_id, pgid-1, &desc_str);
			if (rval == dcb_success && desc_str) {
				sprintf(rbuf, "%01x%02x%s", pgid,
					(unsigned int) strlen(desc_str),
					desc_str);
				if (desc_str)
					free(desc_str);
			}
		}
	}
	return rval;
}

static dcb_result set_bwg_desc(char *port_id, char *ibuf, int ilen)
{
	u8 pgid;
	u8 desc_len = 0;
	int plen;
	dcb_result rval = dcb_success;

	plen = strlen(port_id);
	if (ilen >=  DCB_PORT_OFF + plen + PG_DESC_SET_DLEN) {
		hexstr2bin(ibuf+DCB_PORT_OFF + plen + PG_DESC_LEN,
			&desc_len, sizeof(desc_len));
		pgid = *(ibuf + DCB_PORT_OFF + plen + PG_DESC_PGID) & 0x0f;
	} else {
		printf("ilen[%d] < %d\n", ilen, DCB_PORT_OFF + plen +
			PG_DESC_SET_DLEN);
		return dcb_bad_params;
	}

	if (ilen ==  DCB_PORT_OFF + plen + PG_DESC_SET_DLEN + desc_len) {
		if (pgid < 1 || pgid > 8 || desc_len > 100) {
			printf("pgid = %d, desc_len = %d\n", pgid, desc_len);
			rval = dcb_bad_params;
		} else {
			if (init_cfg()) {
				rval = put_bwg_descrpt(port_id, pgid-1,
					ibuf+DCB_PORT_OFF+plen+PG_DESC_DATA);
				destroy_cfg();
			} else {
				rval = dcb_failed;
			}
		}
	} else {
		printf("ilen[%d] != %d\n", ilen, DCB_PORT_OFF + plen +
			PG_DESC_SET_DLEN + desc_len);
		rval = dcb_bad_params;
	}

	return rval;
}

static void set_protocol_data(feature_protocol_attribs *protocol, char *ibuf,
	int plen)
{
	u8 flag;

	flag = *(ibuf+DCB_PORT_OFF+plen+CFG_ENABLE);
	if (flag == '0' || flag == '1')
		protocol->Enable = flag & 0x01;

	flag = *(ibuf+DCB_PORT_OFF+plen+CFG_ADVERTISE);
	if (flag == '0' || flag == '1')
		protocol->Advertise = flag & 0x01;

	flag = *(ibuf+DCB_PORT_OFF+plen+CFG_WILLING);
	if (flag == '0' || flag == '1')
		protocol->Willing = flag & 0x01;
}

/* rbuf points to correct destination location */
static dcb_result get_cmd_protocol_data(feature_protocol_attribs *protocol,
	u8 cmd, char *rbuf)
{
	dcb_result rval = dcb_success;

	switch(cmd) {
	case CMD_GET_CONFIG:
		sprintf(rbuf, "%1x%1x%1x",
			protocol->Enable,
			protocol->Advertise,
			protocol->Willing);
		break;
	case CMD_GET_OPER:
		sprintf(rbuf, "%02x%02x%02x%1x%1x",
			protocol->Oper_version,
			protocol->Max_version,
			protocol->Error_Flag,
			protocol->OperMode,
			protocol->Syncd);
		break;
	case CMD_GET_PEER:
		if (protocol->TLVPresent == TRUE)
			sprintf(rbuf, "%1x%1x%02x%02x%1x%1x",
				protocol->Enable,
				protocol->Willing,
				protocol->Oper_version,
				protocol->Max_version,
				protocol->Error,
				protocol->dcbx_st);
		else
			rval = dcb_peer_not_present;
		break;
	}
	return rval;
}

static int find_cmd_entry(int cmd)
{
	int i;

	for (i = 0; cmd_tbl[i].cmd != cmd && cmd_tbl[i].cmd != UNKNOWN_CMD; i++)
		;

	return (i);
}

void process_clif_cmd(  struct dcbd_data *dd,
			struct sockaddr_un *from,
			socklen_t fromlen,
			char *ibuf, int ilen, char *rbuf, int *rlen)
{
	int status;

	/* setup minimum command response message
	 * status will be updated at end */
	sprintf(rbuf, "%c%02x", CMD_RESPONSE, dcb_failed);
	*rlen = strlen(rbuf);

	if (ilen < 1) {
		return;
	}

	status = cmd_tbl[find_cmd_entry((int)ibuf[0])].cmd_handler(
		dd, from, fromlen, ibuf, ilen, rbuf+strlen(rbuf));

	/* update status and compute final length */
	rbuf[CLIF_STAT_OFF] = hexlist[ (status & 0x0f0) >> 4 ];
	rbuf[CLIF_STAT_OFF+1] = hexlist[ status & 0x0f ];
	*rlen = strlen(rbuf);
}

static int  handle_dcbx_cmd(u8 cmd, u8 feature, u8 subtype,
				     char *ibuf, int ilen,
				     char *rbuf)
{
	int status = dcb_success;

	if (ilen < DCBX_CFG_OFF)
		return dcb_invalid_cmd;

	/* append standard dcb command response content */
	sprintf(rbuf , "%*.*s", DCBX_CFG_OFF, DCBX_CFG_OFF, ibuf);

	switch(feature) {
	case FEATURE_DCBX:
		if (cmd == CMD_SET_CONFIG) {
			status = set_dcbx_config(ibuf, ilen);
		} else if (cmd != CMD_GET_PEER) {
			status = get_dcbx_config(cmd, rbuf+strlen(rbuf));
		} else {
			status = dcb_invalid_cmd;
		}
		break;
	default:
		break;
	}

	return status;
}

int clif_iface_dcb_cmd(struct dcbd_data *dd,
				     struct sockaddr_un *from,
				     socklen_t fromlen,
				     char *ibuf, int ilen,
				     char *rbuf)
{
	u8 status=dcb_success;
	u8 cmd;
	u8 feature;
	u8 subtype;
	u8 plen;
	char port_id[MAX_U8_BUF];
	pg_attribs pg_data;
	pfc_attribs pfc_data;
	app_attribs app_data;
	llink_attribs llink_data;
	bcn_attribs bcn_data;

	if (hexstr2bin(ibuf+DCB_CMD_OFF, &cmd, sizeof(cmd)) ||
		hexstr2bin(ibuf+DCB_FEATURE_OFF, &feature, sizeof(feature)))
		return dcb_invalid_cmd;

	if (feature == FEATURE_DCBX)
		return handle_dcbx_cmd(cmd, feature, subtype, ibuf, ilen, rbuf);

	if (hexstr2bin(ibuf+DCB_SUBTYPE_OFF, &subtype, sizeof(subtype)) ||
		hexstr2bin(ibuf+DCB_PORTLEN_OFF, &plen, sizeof(plen)))
		return dcb_invalid_cmd;

	if (ilen < DCB_PORT_OFF)
		return dcb_invalid_cmd;
	
	if (ibuf[DCB_VER_OFF] != (CLIF_MSG_VERSION | 0x30)) {
		printf("unsupported client interface message version %x\n",
			ibuf[DCB_VER_OFF]);
		return dcb_invalid_cmd;
	}

	if (ilen < DCB_PORT_OFF+plen) {
		printf("command too short\n");
		return dcb_invalid_cmd;
	}

	/* append standard dcb command response content */
	sprintf(rbuf , "%*.*s", DCB_PORT_OFF+plen, DCB_PORT_OFF+plen, ibuf);

	memcpy(port_id, ibuf+DCB_PORT_OFF, plen);
	port_id[plen] = '\0';

	/* check that interface is present and in DCB state */
	if (feature != FEATURE_DCB && check_port_dcb_mode(port_id) == FALSE)
		return dcb_device_not_found;

	switch(feature) {
	case FEATURE_DCB:
		if (cmd == CMD_SET_CONFIG)
			status = set_dcb_state(port_id, ibuf, ilen);
		else if (cmd == CMD_GET_CONFIG)
			status = get_dcb_state(port_id, rbuf+strlen(rbuf));
		else
			status = dcb_invalid_cmd;
		break;

	case FEATURE_PG:
		if (cmd == CMD_GET_PEER) {
			status = get_peer_pg(port_id, &pg_data);
		} else {
			if (init_cfg()) {
				status = get_pg(port_id, &pg_data);
				destroy_cfg();
			} else {
				status = dcb_failed;
			}
		}

		if (status != dcb_success) {
			printf("error[%d] getting PG data for %s\n",
				status, port_id);
			return status;
		}

		if (cmd == CMD_SET_CONFIG) {
			if (ilen < (DCB_PORT_OFF + plen + CFG_LEN)) {
				printf("set command too short\n");
				status = dcb_invalid_cmd;
			} else {
				set_protocol_data(&pg_data.protocol, ibuf,
					plen);
				status = set_pg_config(&pg_data, port_id, ibuf,
					ilen);
			}
		} else {
			status = get_cmd_protocol_data(&pg_data.protocol, cmd,
					rbuf+strlen(rbuf));
			if (status == dcb_success)
				status = get_pg_data(&pg_data, cmd, port_id,
					pg_data.protocol.dcbx_st,
					rbuf+strlen(rbuf));
		}
		break;

	case FEATURE_PFC:
		if (cmd == CMD_GET_PEER) {
			status = get_peer_pfc(port_id, &pfc_data);
		} else {
			if (init_cfg()) {
				status = get_pfc(port_id, &pfc_data);
				destroy_cfg();
			} else {
				status = dcb_failed;
			}
		}

		if (status != dcb_success) {
			printf("error[%d] getting PFC data for %s\n",
					status, port_id);
			return status;
		}

		if (cmd == CMD_SET_CONFIG) {
			if (ilen < (DCB_PORT_OFF + plen + CFG_LEN)) {
				printf("set command too short\n");
				status = dcb_failed;
			} else {
				set_protocol_data(&pfc_data.protocol, ibuf,
					plen);
				status = set_pfc_config(&pfc_data, port_id,
					ibuf, ilen);
			}
		} else {
			status = get_cmd_protocol_data(&pfc_data.protocol,
				cmd, rbuf+strlen(rbuf));
			if (status == dcb_success)
				status = get_pfc_data(&pfc_data, cmd,
					port_id, pfc_data.protocol.dcbx_st,
					rbuf+strlen(rbuf));
		}
		break;

	case FEATURE_BCN:
		if (cmd == CMD_GET_PEER) {
			status = get_peer_bcn(port_id, &bcn_data);
		} else {
			if (init_cfg()) {
				status = get_bcn(port_id, &bcn_data);
				destroy_cfg();
			} else {
				status = dcb_failed;
			}
		}

		if (status != dcb_success) {
			printf("error[%d] getting BCN data for %s\n",
					status, port_id);
			return status;
		}

		if (cmd == CMD_SET_CONFIG) {
			if (ilen < (DCB_PORT_OFF + plen + CFG_LEN)) {
				printf("set command too short\n");
				status = dcb_failed;
			} else {
				set_protocol_data(&bcn_data.protocol, ibuf,
					plen);
				status = set_bcn_config(&bcn_data, port_id,
					ibuf, ilen);
			}
		} else {
			status = get_cmd_protocol_data(&bcn_data.protocol,
				cmd, rbuf+strlen(rbuf));
			if (status == dcb_success)
				status = get_bcn_data(&bcn_data, cmd,
					port_id, bcn_data.protocol.dcbx_st,
					rbuf+strlen(rbuf));
		}
		break;

	case FEATURE_APP:
		if (cmd == CMD_GET_PEER) {
			status = get_peer_app(port_id, (u32)subtype, &app_data);
		} else {
			if (init_cfg()) {
				status = get_app(port_id, (u32)subtype,
					&app_data);
				destroy_cfg();
			} else {
				status = dcb_failed;
			}
		}

		if (status != dcb_success) {
			printf("error[%d] getting APP data for %s\n", status,
				port_id);
			return status;
		}

		if (cmd == CMD_SET_CONFIG) {
			if (ilen < (DCB_PORT_OFF + plen + CFG_LEN)) {
				printf("set command too short\n");
				status = dcb_failed;
			} else {
				set_protocol_data(&app_data.protocol, ibuf,
					plen);
				status = set_app_config(&app_data, port_id,
					(u32)subtype, ibuf, ilen);
			}
		} else {
			status = get_cmd_protocol_data(&app_data.protocol, cmd,
				rbuf+strlen(rbuf));
			if (status == dcb_success)
				status = get_app_data(&app_data, cmd, port_id,
					subtype, app_data.protocol.dcbx_st,
					rbuf+strlen(rbuf));
		}
		break;

	case FEATURE_LLINK:
		if (cmd == CMD_GET_PEER) {
			status = get_peer_llink(port_id, (u32)subtype,
				&llink_data);
		} else {
			if (init_cfg()) {
				status = get_llink(port_id, (u32)subtype,
					&llink_data);
				destroy_cfg();
			} else {
				status = dcb_failed;
			}
		}

		if (status != dcb_success) {
			printf("error[%d] getting APP data for %s\n", status,
				port_id);
			return status;
		}

		if (cmd == CMD_SET_CONFIG) {
			if (ilen < (DCB_PORT_OFF + plen + CFG_LEN)) {
				printf("set command too short\n");
				status = dcb_failed;
			} else {
				set_protocol_data(&llink_data.protocol, ibuf,
					plen);
				status = set_llink_config(&llink_data, port_id,
					(u32)subtype, ibuf, ilen);
			}
		} else {
			status = get_cmd_protocol_data(&llink_data.protocol,
				cmd, rbuf+strlen(rbuf));
			if (status == dcb_success)
				status = get_llink_data(&llink_data, cmd,
					port_id, subtype,
					llink_data.protocol.dcbx_st,
					rbuf+strlen(rbuf));
		}
		break;

	case FEATURE_PG_DESC:
		if (cmd == CMD_GET_CONFIG) {
			if (init_cfg()) {
				status = get_bwg_desc(port_id, ibuf, ilen,
					rbuf+strlen(rbuf));
				destroy_cfg();
			} else {
				status = dcb_failed;
			}

			if (status != dcb_success) {
				printf("error[%d] getting BWG desc for %s\n",
					status, port_id);
				return status;
			}
		} else if (cmd == CMD_SET_CONFIG) {
			status = set_bwg_desc(port_id, ibuf, ilen);
		}

		break;
	default:
		break;
	}
	return status;
}


/* rbuf points to correct destination location */
static dcb_result get_pg_data(pg_attribs *pg_data, int cmd, char *port_id,
	int dcbx_st, char *rbuf)
{
	int i;
	int status;
	int value;

	switch (cmd) {
	case CMD_GET_CONFIG: /* already have the data object */
	case CMD_GET_PEER:
		break;
	case CMD_GET_OPER:
		status = get_oper_pg(port_id, pg_data);
		if (status != dcb_success) {
			printf("error[%d] getting oper PG data for %s\n",
				status, port_id);
			return status;
		}
		break;
	}

	for (i = 0; i < MAX_USER_PRIORITIES; i++)
		sprintf(rbuf+PG_UP2TC(i), "%1x", pg_data->tx.up[i].tcmap);
	for (i = 0; i < MAX_BANDWIDTH_GROUPS; i++)
		sprintf(rbuf+PG_PG_PCNT(i), "%02x", pg_data->tx.pg_percent[i]);
	for (i = 0; i < MAX_USER_PRIORITIES; i++) {
		if (pg_data->tx.up[i].strict_priority == dcb_link)
			value = LINK_STRICT_PGID;
		else
			value = pg_data->tx.up[i].pgid;
		sprintf(rbuf+PG_UP_PGID(i), "%1X", value);
	}
	for (i = 0; i < MAX_USER_PRIORITIES; i++) {
		if ((cmd != CMD_GET_PEER) ||
			(cmd == CMD_GET_PEER && dcbx_st == dcbx_subtype1))
			sprintf(rbuf+PG_UP_PCNT(i), "%02x",
				pg_data->tx.up[i].percent_of_pg_cap);
		else if (cmd == CMD_GET_PEER && dcbx_st == dcbx_subtype2)
			sprintf(rbuf+PG_UP_PCNT(i), "%c%c",
				CLIF_NOT_SUPPLIED, CLIF_NOT_SUPPLIED);
	}
	for (i = 0; i < MAX_USER_PRIORITIES; i++) {
		if ((cmd != CMD_GET_PEER) ||
			(cmd == CMD_GET_PEER && dcbx_st == dcbx_subtype1)) {
			if (pg_data->tx.up[i].strict_priority == dcb_link)
				value = dcb_none;
			else
				value = pg_data->tx.up[i].strict_priority;
			sprintf(rbuf+PG_UP_STRICT(i), "%1x", value);
		} else if (cmd == CMD_GET_PEER && dcbx_st == dcbx_subtype2) {
			sprintf(rbuf+PG_UP_STRICT(i), "%c", CLIF_NOT_SUPPLIED);
		}
	}

	if ((cmd == CMD_GET_PEER && dcbx_st == dcbx_subtype1) ||
		(cmd == CMD_GET_OPER))
		sprintf(rbuf+PG_UP_NUM_TC, "%c", CLIF_NOT_SUPPLIED);
	else
		sprintf(rbuf+PG_UP_NUM_TC, "%1x", pg_data->num_tcs);

	return dcb_success;
}

/* rbuf points to correct destination location */
static int get_pfc_data(pfc_attribs *pfc_data, int cmd, char *port_id,
	int dcbx_st, char *rbuf)
{
	int i;
	int status;

	switch (cmd) {
	case CMD_GET_CONFIG: /* already have the data object */
		break;
	case CMD_GET_OPER:
		status = get_oper_pfc(port_id, pfc_data);
		if (status != dcb_success) {
			printf("error[%d] getting oper PFC data for %s\n",
				status, port_id);
			return status;
		}
		break;
	case CMD_GET_PEER:
		status = get_peer_pfc(port_id, pfc_data);
		if (status != dcb_success) {
			printf("error[%d] getting peer PFC data for %s\n",
				status, port_id);
			return status;
		}
		break;
	}
	
	for (i = 0; i < MAX_USER_PRIORITIES; i++) {
		sprintf(rbuf+PFC_UP(i), "%1x", pfc_data->admin[i]);
	}

	if ((cmd == CMD_GET_PEER && dcbx_st == dcbx_subtype1) ||
		(cmd == CMD_GET_OPER))
		sprintf(rbuf+PFC_NUM_TC, "%c", CLIF_NOT_SUPPLIED);
	else
		sprintf(rbuf+PFC_NUM_TC, "%1x", pfc_data->num_tcs);

	return dcb_success;
}

/* rbuf points to correct destination location */
static int get_app_data(app_attribs *app_data, int cmd, char *port_id,
		u32 subtype, int dcbx_st, char *rbuf)
{
	int status;
	unsigned int i;

	switch (cmd) {
	case CMD_GET_CONFIG: /* already have the data object */
		break;
	case CMD_GET_OPER:
		status = get_oper_app(port_id, subtype, app_data);
		if (status != dcb_success) {
			printf("error[%d] getting oper App data for %s\n",
				status, port_id);
			return status;
		}
		break;
	case CMD_GET_PEER:
		status = get_peer_app(port_id, subtype, app_data);
		if (status != dcb_success) {
			printf("error[%d] getting peer app data for %s\n",
				status, port_id);
			return status;
		}
		break;
	}
	
	sprintf(rbuf+APP_LEN, "%02x", 2*app_data->Length);
	
	for (i=0; i<app_data->Length; i++)
		sprintf(rbuf+APP_DATA+2*i, "%02x", *(app_data->AppData+i));

	return dcb_success;
}

/* rbuf points to correct destination location */
static int get_llink_data(llink_attribs *llink_data, int cmd, char *port_id,
		u32 subtype, int dcbx_st, char *rbuf)
{
	int status;

	switch (cmd) {
	case CMD_GET_CONFIG: /* already have the data object */
		break;
	case CMD_GET_OPER:
		status = get_oper_llink(port_id, subtype, llink_data);
		if (status != dcb_success) {
			printf("error[%d] getting oper logical link data "
				"for %s\n", status, port_id);
			return status;
		}
		break;
	case CMD_GET_PEER:
		status = get_peer_llink(port_id, subtype, llink_data);
		if (status != dcb_success) {
			printf("error[%d] getting peer logical link data "
				"for %s\n", status, port_id);
			return status;
		}
		break;
	}
	
	sprintf(rbuf+LLINK_STATUS, "%1x", llink_data->llink.llink_status);

	return dcb_success;
}

static int get_bcn_data(bcn_attribs *bcn_data, int cmd, char *port_id, 
						int dcbx_st, char *rbuf)
{
	feature_support dcb_support;
	int i;
	int status;
	float f;
	u8 bitmap;

	if (get_dcb_support(port_id, &dcb_support) != dcb_success)
		return dcb_failed;

	if (!dcb_support.bcn && cmd != CMD_GET_PEER)
		return dcb_not_capable;

	switch (cmd) {
	case CMD_GET_CONFIG: /* already have the data object */
		break;
	case CMD_GET_OPER: 
		status = get_oper_bcn(port_id, bcn_data);
		if (status != dcb_success) {
			printf("error[%d] getting oper BCN data for %s\n",
				status, port_id);
			return status;
		}
		break;
	case CMD_GET_PEER: /* need it?*/
		status = get_peer_bcn(port_id, bcn_data);
		if (status != dcb_success) {
			printf("error[%d] getting peer BCN data for %s\n",
				status, port_id);
			return status;
		}
		break;
	}

	for (i = 0; i < MAX_USER_PRIORITIES; i++) {
		sprintf(rbuf+BCN_UP_RP(i), "%1x", 
				bcn_data->bcn.up_settings[i].rp_admin);
	}

	if (cmd == CMD_GET_PEER || cmd == CMD_GET_OPER) {
		bitmap = 0;
		for (i = 0; i < MAX_USER_PRIORITIES; i++)
			if (bcn_data->bcn.up_settings[i].rp_oper)
				bitmap |= (1<<i);
		sprintf(rbuf+BCN_UP_RP_OP, "%02x", bitmap);
		bitmap = 0;
		for (i = 0; i < MAX_USER_PRIORITIES; i++)
			if (bcn_data->bcn.up_settings[i].rem_tag_oper)
				bitmap |= (1<<i);
		sprintf(rbuf+BCN_UP_RP_RTAG_OP, "%02x", bitmap);
	} else {
		/* two CLIF_NOT_SUPPLIED per field to maintain
		 * offsets.  */
		sprintf(rbuf+BCN_UP_RP_OP, "%c%c%c%c",
			CLIF_NOT_SUPPLIED, CLIF_NOT_SUPPLIED,
			CLIF_NOT_SUPPLIED, CLIF_NOT_SUPPLIED);
	}

	if (cmd == CMD_GET_PEER) {
		bitmap = 0;
		for (i = 0; i < MAX_USER_PRIORITIES; i++)
			if (bcn_data->bcn.up_settings[i].cp_admin)
				bitmap |= (1<<i);
		sprintf(rbuf+BCN_UP_CP, "%02x", bitmap);
	} else {
		/* two CLIF_NOT_SUPPLIED per field to maintain
		 * offsets.  */
		sprintf(rbuf+BCN_UP_CP, "%c%c",
			CLIF_NOT_SUPPLIED, CLIF_NOT_SUPPLIED);
	}

	for (i = 0; i < BCN_ADDR_OPTION_LEN; i++) {
		sprintf(rbuf+BCN_BCNA(i), "%02x", 
				bcn_data->bcn.bcna[i]);
	}

	f = bcn_data->bcn.rp_alpha;
	sprintf(rbuf+BCN_ALPHA,"%08x", *((u32 *)(&f)));
	f = bcn_data->bcn.rp_beta;
	sprintf(rbuf+BCN_BETA,"%08x",  *((u32 *)(&f)));
	f = bcn_data->bcn.rp_gd;
	sprintf(rbuf+BCN_GD,"%08x",    *((u32 *)(&f)));
	f = bcn_data->bcn.rp_gi;
	sprintf(rbuf+BCN_GI,"%08x",    *((u32 *)(&f)));
	sprintf(rbuf+BCN_TMAX, "%08x", bcn_data->bcn.rp_tmax);
	sprintf(rbuf+BCN_RI,   "%08x", bcn_data->bcn.rp_ri);
	sprintf(rbuf+BCN_TD,   "%04x", bcn_data->bcn.rp_td);
	sprintf(rbuf+BCN_RMIN, "%04x", bcn_data->bcn.rp_rmin);
	sprintf(rbuf+BCN_W,    "%02x", bcn_data->bcn.rp_w);
	sprintf(rbuf+BCN_RD,   "%02x", bcn_data->bcn.rp_rd);
	sprintf(rbuf+BCN_RU,   "%02x", bcn_data->bcn.rp_ru);
	sprintf(rbuf+BCN_WRTT, "%02x", bcn_data->bcn.rp_wrtt);
	if (cmd == CMD_GET_PEER) {
		sprintf(rbuf+BCN_CP_SF, "%08x", bcn_data->bcn.cp_sf);
	} else {
		/* last field in the message, just send one
		 * CLIF_NOT_SUPPLIED to indicate it's not present.
		 */
		sprintf(rbuf+BCN_CP_SF, "%c",
			CLIF_NOT_SUPPLIED);
	}

	return dcb_success;
}

static int set_pg_config(pg_attribs *pg_data, char *port_id, char *ibuf,
	int ilen)
{
	full_dcb_attrib_ptrs dcb_data;
	u8 flag;
	dcb_result status = dcb_success;
	int i;
	int plen;
	int off;
	u8 used[MAX_BANDWIDTH_GROUPS];
	u8 uppcts_changed = FALSE;

	plen=strlen(port_id);
	off = DCB_PORT_OFF + plen + CFG_LEN;

	if (ilen == (off + CFG_PG_DLEN)) {
		for (i = 0; i < MAX_USER_PRIORITIES; i++) {
			flag = *(ibuf+off+PG_UP2TC(i));
			if (flag == CLIF_NOT_SUPPLIED)
				continue;
			pg_data->tx.up[i].tcmap = flag & 0x07;
			pg_data->rx.up[i].tcmap = flag & 0x07;
		}
		memset(used, FALSE, sizeof(used));
		for (i = 0; i < MAX_BANDWIDTH_GROUPS; i++) {
			if (*(ibuf+off+PG_PG_PCNT(i)) == CLIF_NOT_SUPPLIED)
				continue;
			if (hexstr2bin(ibuf+off+PG_PG_PCNT(i),
				&flag, sizeof(flag)))
				return dcb_bad_params;
			pg_data->tx.pg_percent[i] = flag;
			pg_data->rx.pg_percent[i] = flag;
		}
		for (i = 0; i < MAX_USER_PRIORITIES; i++) {
			flag = *(ibuf+off+PG_UP_PGID(i));
			if (flag == CLIF_NOT_SUPPLIED) {
				if (pg_data->tx.up[i].strict_priority ==
					dcb_link)
					flag = LINK_STRICT_PGID;
				else
					flag = pg_data->tx.up[i].pgid;
			} else {
				if (flag == hexlist[LINK_STRICT_PGID] ||
					flag == hexlistcaps[LINK_STRICT_PGID])
					flag = LINK_STRICT_PGID;
				else
					flag = flag & 0x0f;
				pg_data->tx.up[i].pgid = flag;
				pg_data->rx.up[i].pgid = flag;
			}

			/* keep track of which PGID's in the range of
			 * 0-7 are being used (not counting pre-existing
			 * link strict PGID).
			*/
			if (flag < MAX_BANDWIDTH_GROUPS)
				used[flag] = TRUE;
		}
		for (i = 0; i < MAX_USER_PRIORITIES; i++) {
			if (*(ibuf+off+PG_UP_PCNT(i)) == CLIF_NOT_SUPPLIED)
				continue;
			uppcts_changed = TRUE;
			if (hexstr2bin(ibuf+off+PG_UP_PCNT(i),
				&flag, sizeof(flag)))
				return dcb_bad_params;
			pg_data->tx.up[i].percent_of_pg_cap = flag;
			pg_data->rx.up[i].percent_of_pg_cap = flag;
		}
		for (i = 0; i < MAX_USER_PRIORITIES; i++) {
			flag = *(ibuf+off+PG_UP_STRICT(i));
			if (flag == CLIF_NOT_SUPPLIED)
				continue;
			/* only set or clear the group strict bit.
			 * the link strict bit will be handled later.
			*/
			flag = flag & 0x01;
			if (flag) {
				pg_data->tx.up[i].strict_priority |= flag;
				pg_data->rx.up[i].strict_priority |= flag;
			} else {
				pg_data->tx.up[i].strict_priority &= ~dcb_group;
				pg_data->rx.up[i].strict_priority &= ~dcb_group;
			}
		}

		/* find the first unused PGID in range 0-7 */
		for (i = 0; i < MAX_BANDWIDTH_GROUPS; i++)
			if (!used[i])
				break;

		/* The end goal here is to have all of the user priorities
		 * assigned to the link strict PGID to use the lowest
		 * unused value in the range 0-7.  The strict_priority field
		 * is set to 'dcb_link' as well for the link strict PGID.
		 */
		flag = i;
		for (i = 0; i < MAX_USER_PRIORITIES; i++) {
			if (pg_data->tx.up[i].pgid == LINK_STRICT_PGID ||
				(!used[pg_data->tx.up[i].pgid] &&
				pg_data->tx.up[i].strict_priority & dcb_link)) {
				pg_data->tx.up[i].pgid = flag;
				pg_data->rx.up[i].pgid = flag;
				pg_data->tx.up[i].strict_priority = dcb_link;
				pg_data->rx.up[i].strict_priority = dcb_link;
			} else {
				pg_data->tx.up[i].strict_priority &= ~dcb_link;
				pg_data->rx.up[i].strict_priority &= ~dcb_link;
			}
		}
	} else if (ilen != off) {
		printf("error - setcommand has invalid argument length\n");
		return dcb_bad_params;
	}

	memset((void *)&dcb_data, 0, sizeof(dcb_data));
	dcb_data.pg = pg_data;
	status = dcb_check_config(&dcb_data);

	/* if the rule checking fails and client did not supply
	 * user priority percentages, then compute new percentages
	 * and try one more time.
	*/
	if (status == dcb_bad_params && !uppcts_changed) {
		rebalance_uppcts(pg_data);
		status = dcb_check_config(&dcb_data);
	}

	if (status != dcb_success) {
		printf("invalid DCB settings\n");
		return status;
	}

	if (init_cfg()) {
		status = put_pg(port_id, pg_data);
		destroy_cfg();
	} else {
		status = dcb_failed;
	}
	if (status != dcb_success)
		printf("error[%d] setting PG data for %s\n", status, port_id);
	

	return status;
}

static int set_pfc_config(pfc_attribs *pfc_data, char *port_id, char *ibuf,
	int ilen)
{
	u8 flag;
	dcb_result status = dcb_success;
	int i;
	int off;
	int plen;
	
	plen=strlen(port_id);
	off = DCB_PORT_OFF + plen + CFG_LEN;

	if (ilen == (off + CFG_PFC_DLEN)) {
		for (i = 0; i < MAX_USER_PRIORITIES; i++) {
			flag = (*(ibuf+off+PFC_UP(i))) & 0x07;
			if (flag == CLIF_NOT_SUPPLIED)
				continue;
			if (flag)
				pfc_data->admin[i] = pfc_enabled;
			else
				pfc_data->admin[i] = pfc_disabled;
		}
	} else if (ilen != off) {
		/* at least needs to include the protocol settings */
		printf("error - setcommand has invalid argument length\n");
		return dcb_failed;
	}

	if (init_cfg()) {
		status = put_pfc(port_id, pfc_data);
		destroy_cfg();
	} else {
		status = dcb_failed;
	}
	if (status != dcb_success)
		printf("error[%d] setting PFC data for %s\n", status, port_id);

	return status;
}

static int set_app_config(app_attribs *app_data, char *port_id, u32 subtype,
		char *ibuf, int ilen)
{
	dcb_result status = dcb_success;
	int off;
	int plen;
	u8 applen = 0;
	
	plen=strlen(port_id);
	off = DCB_PORT_OFF + plen + CFG_LEN;

	hexstr2bin(ibuf+off+APP_LEN, &applen, sizeof(applen));

	if (ilen == (off + APP_DATA + applen)) {
		app_data->Length = applen/2;
		if (hexstr2bin(ibuf+off+APP_DATA, app_data->AppData, applen/2))
			return dcb_bad_params;
	} else if (ilen != off) {
		/* at least needs to include the protocol settings */
		printf("error - setcommand has invalid argument length\n");
		return dcb_failed;
	}

	if (init_cfg()) {
		status = put_app(port_id, subtype, app_data);
		destroy_cfg();
	} else {
		status = dcb_failed;
	}
	if (status != dcb_success)
		printf("error[%d] setting APP data for %s\n", status, port_id);

	return status;
}

static int set_llink_config(llink_attribs *llink_data, char *port_id,
		u32 subtype, char *ibuf, int ilen)
{
	dcb_result status = dcb_success;
	int off;
	int value;
	
	off = DCB_PORT_OFF + strlen(port_id) + CFG_LEN;

	if (ilen != off + LLINK_DLEN) {
		/* at least needs to include the protocol settings */
		printf("error - setcommand has invalid argument length\n");
		return dcb_failed;
	}

	value = *(ibuf+off+LLINK_STATUS);
	if (value != CLIF_NOT_SUPPLIED)
		llink_data->llink.llink_status = value & 0x0f;

	if (init_cfg()) {
		status = put_llink(port_id, subtype, llink_data);
		destroy_cfg();
	} else {
		status = dcb_failed;
	}

	if (status != dcb_success)
		printf("error[%d] setting logical link data for %s\n",
			status, port_id);

	return status;
}

static int set_bcn_config(bcn_attribs *bcn_data, char *port_id, char *ibuf, 
						  int ilen)
{
	feature_support dcb_support;
	bcn_attribs sav_bcn_data;
	u8 flag;
	dcb_result status = dcb_success;
	int i;
	int off;
	int plen;
	u32 n;

	if (get_dcb_support(port_id, &dcb_support) != dcb_success)
		return dcb_failed;

	if (!dcb_support.bcn)
		return dcb_not_capable;
	
	plen=strlen(port_id);
	off = DCB_PORT_OFF + plen + CFG_LEN;

	if (ilen == off)
		goto save_bcn; /* only DCBX parameters were supplied) */
	else if (off + MAX_USER_PRIORITIES > ilen)
		return dcb_failed;

	/* save copy of bcn_data in case a parse error occurs */
	memcpy((void *)&sav_bcn_data, (void *)bcn_data, sizeof(bcn_attribs));

	for (i = 0; i < MAX_USER_PRIORITIES; i++) {
		flag = *(ibuf+off+BCN_UP_RP(i));
		if (flag == CLIF_NOT_SUPPLIED)
			continue;
		if (flag & 0x07)
			bcn_data->bcn.up_settings[i].rp_admin = 1; 
		else
			bcn_data->bcn.up_settings[i].rp_admin = 0;
	}
	off += MAX_USER_PRIORITIES;

	/*for (i = 0; i < BCN_ADDR_OPTION_LEN; i++) {
		flag = *(ibuf+off+BCN_BCNA(i));
		if (flag == CLIF_NOT_SUPPLIED)
			continue;
		hexstr2bin(ibuf+off, (u8*)&n, sizeof(u8));
		bcn_data->bcn.bcna[i] = n;		
	}
	off += BCN_ADDR_OPTION_LEN;*/

	if (ibuf[off] != CLIF_NOT_SUPPLIED) {
		hexstr2bin(ibuf+off, (u8*)&n, sizeof(u32));
		n = ntohl(n);
		bcn_data->bcn.rp_alpha = (double) *((float *)&n);
		off += 8;
	} else {
		off++;
	}

	if (ibuf[off] != CLIF_NOT_SUPPLIED) {
		hexstr2bin(ibuf+off, (u8*)&n, sizeof(u32));
		n = ntohl(n);
		bcn_data->bcn.rp_beta = (double) *((float *)&n);
		off += 8;
	} else {
		off++;
	}

	if (ibuf[off] != CLIF_NOT_SUPPLIED) {
		hexstr2bin(ibuf+off, (u8*)&n, sizeof(u32));
		n = ntohl(n);
		bcn_data->bcn.rp_gd = (double) *((float *)&n);
		off += 8;
	} else {
		off++;
	}

	if (ibuf[off] != CLIF_NOT_SUPPLIED) {
		hexstr2bin(ibuf+off, (u8*)&n, sizeof(u32));
		n = ntohl(n);
		bcn_data->bcn.rp_gi = (double) *((float *)&n);
		off += 8;
	} else {
		off++;
	}

	if (ibuf[off] != CLIF_NOT_SUPPLIED) {
		hexstr2bin(ibuf+off, (u8*)&n, sizeof(u32));
		bcn_data->bcn.rp_tmax = ntohl(n);
		off += 8;
	} else {
		off++;
	}

	if (ibuf[off] != CLIF_NOT_SUPPLIED) {
		hexstr2bin(ibuf+off, (u8*)&n, sizeof(u32));
		bcn_data->bcn.rp_ri = ntohl(n);
		off += 8;
	} else {
		off++;
	}

	if (ibuf[off] != CLIF_NOT_SUPPLIED) {
		hexstr2bin(ibuf+off, (u8*)&n, sizeof(u16));
		bcn_data->bcn.rp_td = ntohs(n);
		off += 4;
	} else {
		off++;
	}

	if (ibuf[off] != CLIF_NOT_SUPPLIED) {
		hexstr2bin(ibuf+off, (u8*)&n, sizeof(u16));
		bcn_data->bcn.rp_rmin = ntohs(n);
		off += 4;
	} else {
		off++;
	}

	if (ibuf[off] != CLIF_NOT_SUPPLIED) {
		hexstr2bin(ibuf+off, (u8*)&n, sizeof(u8));
		bcn_data->bcn.rp_w = n;
		off += 2;
	} else {
		off++;
	}

	if (ibuf[off] != CLIF_NOT_SUPPLIED) {
		hexstr2bin(ibuf+off, (u8*)&n, sizeof(u8));
		bcn_data->bcn.rp_rd = n;
		off += 2;
	} else {
		off++;
	}

	if (ibuf[off] != CLIF_NOT_SUPPLIED) {
		hexstr2bin(ibuf+off, (u8*)&n, sizeof(u8));
		bcn_data->bcn.rp_ru = n;
		off += 2;
	} else {
		off++;
	}

	if (ibuf[off] != CLIF_NOT_SUPPLIED) {
		hexstr2bin(ibuf+off, (u8*)&n, sizeof(u8));
		bcn_data->bcn.rp_wrtt = n;
		off += 2;
	} else {
		off++;
	}

	/* if message length doesn't match up, then data in the
	 * bcn_data structure is no good. Restore copy.
	*/
	if (off != ilen) {
		printf("invalid client interface message for BCN\n");
		return dcb_failed;
		memcpy((void *)bcn_data, (void *)&sav_bcn_data,
			sizeof(bcn_cfg));
	}

save_bcn:
	if (init_cfg()) {
		/*TODO: delete comment in shared code */
		status = put_bcn(port_id, bcn_data); 
		destroy_cfg();
	} else {
		status = dcb_failed;
	}
	if (status != dcb_success)
		printf("error[%d] setting BCN data for %s\n", status, port_id);

	return status;
}

void pfc_event(char *dn, u32 events)
{
	char ebuf[512];

	if (snprintf(ebuf, sizeof(ebuf), "%01x%02x%s%02x00%1x%1x",
		CLIF_EV_VERSION, (unsigned int)strlen(dn), dn, FEATURE_PFC,
		!!(events & EVENT_OPERMODE),
		!!(events & EVENT_OPERATTR)) < (int)sizeof(ebuf))
		send_event(MSG_DCB, ebuf);
}

void pg_event(char *dn, u32 events)
{
	char ebuf[512];

	if (snprintf(ebuf, sizeof(ebuf), "%01x%02x%s%02x00%1x%1x",
		CLIF_EV_VERSION, (unsigned int)strlen(dn), dn, FEATURE_PG,
		!!(events & EVENT_OPERMODE),
		!!(events & EVENT_OPERATTR)) < (int)sizeof(ebuf))
		send_event(MSG_DCB, ebuf);
}

void app_event(char *dn, u32 subtype, u32 events)
{
	char ebuf[512];

	if (snprintf(ebuf, sizeof(ebuf), "%01x%02x%s%02x%02x%1x%1x",
		CLIF_EV_VERSION, (unsigned int)strlen(dn), dn, FEATURE_APP,
		subtype,
		!!(events & EVENT_OPERMODE),
		!!(events & EVENT_OPERATTR)) < (int)sizeof(ebuf))
		send_event(MSG_DCB, ebuf);
}

void llink_event(char *dn, u32 subtype, u32 events)
{
	char ebuf[512];

	if (snprintf(ebuf, sizeof(ebuf), "%01x%02x%s%02x%02x%1x%1x",
		CLIF_EV_VERSION, (unsigned int)strlen(dn), dn, FEATURE_LLINK,
		subtype,
		!!(events & EVENT_OPERMODE),
		!!(events & EVENT_OPERATTR)) < (int)sizeof(ebuf))
		send_event(MSG_DCB, ebuf);
}

void bcn_event(char *dn, u32 events)
{
/*TODO: similar to other xxx_event, which could call a single function */
/*TODO: delete //TODO in shared code */
	char ebuf[512];

	if (snprintf(ebuf, sizeof(ebuf), "%01x%02x%s%02x00%1x%1x",
		CLIF_EV_VERSION, (unsigned int)strlen(dn), dn, FEATURE_BCN,
		!!(events & EVENT_OPERMODE),
		!!(events & EVENT_OPERATTR)) < (int)sizeof(ebuf))
		send_event(MSG_DCB, ebuf);
}
